from typing import TYPE_CHECKING, Optional, Tuple

from .exceptions import get_status_code_error
from .specs.openapi.checks import (
    content_type_conformance,
    response_headers_conformance,
    response_schema_conformance,
    status_code_conformance,
)
from .utils import GenericResponse

if TYPE_CHECKING:
    from .models import Case, CheckFunction


def not_a_server_error(response: GenericResponse, case: "Case") -> Optional[bool]:  # pylint: disable=useless-return
    """A check to verify that the response is not a server-side error."""
    if response.status_code >= 500:
        exc_class = get_status_code_error(response.status_code)
        raise exc_class(f"Received a response with 5xx status code: {response.status_code}")
    return None


DEFAULT_CHECKS: Tuple["CheckFunction", ...] = (not_a_server_error,)
OPTIONAL_CHECKS = (
    status_code_conformance,
    content_type_conformance,
    response_headers_conformance,
    response_schema_conformance,
)
ALL_CHECKS: Tuple["CheckFunction", ...] = DEFAULT_CHECKS + OPTIONAL_CHECKS



import click

from ...runner import events
from ..context import ExecutionContext
from ..handlers import EventHandler
from . import default


def handle_before_execution(context: ExecutionContext, event: events.BeforeExecution) -> None:
    if event.recursion_level > 0:
        context.operations_count += 1  # type: ignore


def handle_after_execution(context: ExecutionContext, event: events.AfterExecution) -> None:
    context.operations_processed += 1
    context.results.append(event.result)
    context.hypothesis_output.extend(event.hypothesis_output)
    default.display_execution_result(context, event)


class ShortOutputStyleHandler(EventHandler):
    def handle_event(self, context: ExecutionContext, event: events.ExecutionEvent) -> None:
        """Short output style shows single symbols in the progress bar.

        Otherwise, identical to the default output style.
        """
        if isinstance(event, events.Initialized):
            default.handle_initialized(context, event)
        if isinstance(event, events.BeforeExecution):
            handle_before_execution(context, event)
        if isinstance(event, events.AfterExecution):
            handle_after_execution(context, event)
        if isinstance(event, events.Finished):
            if context.operations_count == context.operations_processed:
                click.echo()
            default.handle_finished(context, event)
        if isinstance(event, events.Interrupted):
            default.handle_interrupted(context, event)
        if isinstance(event, events.InternalError):
            default.handle_internal_error(context, event)



from hashlib import sha1
from json import JSONDecodeError
from typing import Any, Callable, Dict, NoReturn, Optional, Type, Union

import attr
from jsonschema import ValidationError

from .constants import SERIALIZERS_SUGGESTION_MESSAGE
from .utils import GenericResponse


class CheckFailed(AssertionError):
    """Custom error type to distinguish from arbitrary AssertionError that may happen in the dependent libraries."""

    __module__ = "builtins"


CACHE: Dict[Union[str, int], Type[CheckFailed]] = {}


def get_exception(name: str) -> Type[CheckFailed]:
    """Create a new exception class with provided name or fetch one from the cache."""
    if name in CACHE:
        exception_class = CACHE[name]
    else:
        exception_class = type(name, (CheckFailed,), {})
        exception_class.__qualname__ = CheckFailed.__name__
        exception_class.__name__ = CheckFailed.__name__
        CACHE[name] = exception_class
    return exception_class


def _get_hashed_exception(prefix: str, message: str) -> Type[CheckFailed]:
    """Give different exceptions for different error messages."""
    messages_digest = sha1(message.encode("utf-8")).hexdigest()
    name = f"{prefix}{messages_digest}"
    return get_exception(name)


def get_grouped_exception(prefix: str, *exceptions: AssertionError) -> Type[CheckFailed]:
    # The prefix is needed to distinguish multiple operations with the same error messages
    # that are coming from different operations
    messages = [exception.args[0] for exception in exceptions]
    message = "".join(messages)
    return _get_hashed_exception("GroupedException", f"{prefix}{message}")


def get_status_code_error(status_code: int) -> Type[CheckFailed]:
    """Return new exception for an unexpected status code."""
    name = f"StatusCodeError{status_code}"
    return get_exception(name)


def get_response_type_error(expected: str, received: str) -> Type[CheckFailed]:
    """Return new exception for an unexpected response type."""
    name = f"SchemaValidationError{expected}_{received}"
    return get_exception(name)


def get_malformed_media_type_error(media_type: str) -> Type[CheckFailed]:
    name = f"MalformedMediaType{media_type}"
    return get_exception(name)


def get_missing_content_type_error() -> Type[CheckFailed]:
    """Return new exception for a missing Content-Type header."""
    return get_exception("MissingContentTypeError")


def get_schema_validation_error(exception: ValidationError) -> Type[CheckFailed]:
    """Return new exception for schema validation error."""
    return _get_hashed_exception("SchemaValidationError", str(exception))


def get_response_parsing_error(exception: JSONDecodeError) -> Type[CheckFailed]:
    """Return new exception for response parsing error."""
    return _get_hashed_exception("ResponseParsingError", str(exception))


def get_headers_error(message: str) -> Type[CheckFailed]:
    """Return new exception for missing headers."""
    return _get_hashed_exception("MissingHeadersError", message)


@attr.s(slots=True)
class InvalidSchema(Exception):
    """Schema associated with an API operation contains an error."""

    __module__ = "builtins"
    message: Optional[str] = attr.ib(default=None)
    path: Optional[str] = attr.ib(default=None)
    method: Optional[str] = attr.ib(default=None)
    full_path: Optional[str] = attr.ib(default=None)

    def as_failing_test_function(self) -> Callable:
        """Create a test function that will fail.

        This approach allows us to use default pytest reporting style for operation-level schema errors.
        """

        def actual_test(*args: Any, **kwargs: Any) -> NoReturn:
            __tracebackhide__ = True  # pylint: disable=unused-variable
            raise self

        return actual_test


class NonCheckError(Exception):
    """An error happened in side the runner, but is not related to failed checks.

    Used primarily to not let Hypothesis to consider the test as flaky or detect multiple failures as we handle it
    on our side.
    """

    __module__ = "builtins"


SERIALIZATION_NOT_POSSIBLE_MESSAGE = (
    f"Schemathesis can't serialize data to any of the defined media types: {{}} \n{SERIALIZERS_SUGGESTION_MESSAGE}"
)


class SerializationNotPossible(Exception):
    """Not possible to serialize to any of the media types defined for some API operation.

    Usually, there is still `application/json` along with less common ones, but this error happens when there is no
    media type that Schemathesis knows how to serialize data to.
    """

    __module__ = "builtins"

    @classmethod
    def from_media_types(cls, *media_types: str) -> "SerializationNotPossible":
        return cls(SERIALIZATION_NOT_POSSIBLE_MESSAGE.format(", ".join(media_types)))


class InvalidRegularExpression(Exception):
    __module__ = "builtins"


@attr.s  # pragma: no mutate
class HTTPError(Exception):
    response: GenericResponse = attr.ib()  # pragma: no mutate
    url: str = attr.ib()  # pragma: no mutate



from functools import lru_cache
from typing import Generator, List, Union

from . import lexer, nodes
from .errors import RuntimeExpressionError, UnknownToken


@lru_cache()  # pragma: no mutate
def parse(expr: str) -> List[nodes.Node]:
    """Parse lexical tokens into concrete expression nodes."""
    return list(_parse(expr))


def _parse(expr: str) -> Generator[nodes.Node, None, None]:
    tokens = lexer.tokenize(expr)
    brackets_stack: List[str] = []
    for token in tokens:
        if token.is_string:
            yield nodes.String(token.value)
        elif token.is_variable:
            yield from _parse_variable(tokens, token, expr)
        elif token.is_left_bracket:
            if brackets_stack:
                raise RuntimeExpressionError("Nested embedded expressions are not allowed")
            brackets_stack.append("{")
        elif token.is_right_bracket:
            if not brackets_stack:
                raise RuntimeExpressionError("Unmatched bracket")
            brackets_stack.pop()
    if brackets_stack:
        raise RuntimeExpressionError("Unmatched bracket")


def _parse_variable(tokens: lexer.TokenGenerator, token: lexer.Token, expr: str) -> Generator[nodes.Node, None, None]:
    if token.value == nodes.NodeType.URL.value:
        yield nodes.URL()
    elif token.value == nodes.NodeType.METHOD.value:
        yield nodes.Method()
    elif token.value == nodes.NodeType.STATUS_CODE.value:
        yield nodes.StatusCode()
    elif token.value == nodes.NodeType.REQUEST.value:
        yield _parse_request(tokens, expr)
    elif token.value == nodes.NodeType.RESPONSE.value:
        yield _parse_response(tokens, expr)
    else:
        raise UnknownToken(token.value)


def _parse_request(tokens: lexer.TokenGenerator, expr: str) -> Union[nodes.BodyRequest, nodes.NonBodyRequest]:
    skip_dot(tokens, "$request")
    location = next(tokens)
    if location.value in ("query", "path", "header"):
        skip_dot(tokens, f"$request.{location.value}")
        parameter = take_string(tokens, expr)
        return nodes.NonBodyRequest(location.value, parameter)
    if location.value == "body":
        try:
            token = next(tokens)
            if token.is_pointer:
                return nodes.BodyRequest(token.value)
        except StopIteration:
            return nodes.BodyRequest()
    raise RuntimeExpressionError(f"Invalid expression: {expr}")


def _parse_response(tokens: lexer.TokenGenerator, expr: str) -> Union[nodes.HeaderResponse, nodes.BodyResponse]:
    skip_dot(tokens, "$response")
    location = next(tokens)
    if location.value == "header":
        skip_dot(tokens, f"$response.{location.value}")
        parameter = take_string(tokens, expr)
        return nodes.HeaderResponse(parameter)
    if location.value == "body":
        try:
            token = next(tokens)
            if token.is_pointer:
                return nodes.BodyResponse(token.value)
        except StopIteration:
            return nodes.BodyResponse()
    raise RuntimeExpressionError(f"Invalid expression: {expr}")


def skip_dot(tokens: lexer.TokenGenerator, name: str) -> None:
    token = next(tokens)
    if not token.is_dot:
        raise RuntimeExpressionError(f"`{name}` expression should be followed by a dot (`.`). Got: {token.value}")


def take_string(tokens: lexer.TokenGenerator, expr: str) -> str:
    parameter = next(tokens)
    if not parameter.is_string:
        raise RuntimeExpressionError(f"Invalid expression: {expr}")
    return parameter.value



from typing import Any, Dict

from ..hooks import HookContext, register, unregister
from ..utils import traverse_schema


def install() -> None:
    register(before_load_schema)


def uninstall() -> None:
    unregister(before_load_schema)


def before_load_schema(context: HookContext, schema: Dict[str, Any]) -> None:
    traverse_schema(schema, _handle_boundaries)


def _handle_boundaries(schema: Dict[str, Any]) -> Dict[str, Any]:
    """Convert Draft 7 keywords to Draft 4 compatible versions.

    FastAPI uses ``pydantic``, which generates Draft 7 compatible schemas.
    """
    for boundary_name, boundary_exclusive_name in (("maximum", "exclusiveMaximum"), ("minimum", "exclusiveMinimum")):
        value = schema.get(boundary_exclusive_name)
        # `bool` check is needed, since in Python `True` is an instance of `int`
        if isinstance(value, (int, float)) and not isinstance(value, bool):
            schema[boundary_exclusive_name] = True
            schema[boundary_name] = value
    return schema



from typing import Any, Callable, Dict, Generator, Iterable, List, Optional, Tuple, Union
from urllib.parse import urlparse

import hypothesis.errors
from starlette.applications import Starlette

from .. import fixups as _fixups
from .. import loaders
from ..checks import DEFAULT_CHECKS
from ..constants import DEFAULT_DATA_GENERATION_METHODS, DEFAULT_STATEFUL_RECURSION_LIMIT, DataGenerationMethod
from ..models import CheckFunction
from ..schemas import BaseSchema
from ..stateful import Stateful
from ..targets import DEFAULT_TARGETS, Target
from ..types import Filter, NotSet, RawAuth
from ..utils import dict_not_none_values, dict_true_values, file_exists, get_requests_auth, import_app
from . import events
from .impl import (
    BaseRunner,
    SingleThreadASGIRunner,
    SingleThreadRunner,
    SingleThreadWSGIRunner,
    ThreadPoolASGIRunner,
    ThreadPoolRunner,
    ThreadPoolWSGIRunner,
)


def prepare(
    schema_uri: Union[str, Dict[str, Any]],
    *,
    # Runtime behavior
    checks: Iterable[CheckFunction] = DEFAULT_CHECKS,
    data_generation_methods: Tuple[DataGenerationMethod, ...] = DEFAULT_DATA_GENERATION_METHODS,
    max_response_time: Optional[int] = None,
    targets: Iterable[Target] = DEFAULT_TARGETS,
    workers_num: int = 1,
    seed: Optional[int] = None,
    exit_first: bool = False,
    dry_run: bool = False,
    store_interactions: bool = False,
    fixups: Iterable[str] = (),
    stateful: Optional[Stateful] = None,
    stateful_recursion_limit: int = DEFAULT_STATEFUL_RECURSION_LIMIT,
    # Schema loading
    loader: Callable = loaders.from_uri,
    base_url: Optional[str] = None,
    auth: Optional[Tuple[str, str]] = None,
    auth_type: Optional[str] = None,
    headers: Optional[Dict[str, str]] = None,
    request_timeout: Optional[int] = None,
    request_tls_verify: Union[bool, str] = True,
    endpoint: Optional[Filter] = None,
    method: Optional[Filter] = None,
    tag: Optional[Filter] = None,
    operation_id: Optional[Filter] = None,
    app: Optional[str] = None,
    validate_schema: bool = True,
    skip_deprecated_operations: bool = False,
    force_schema_version: Optional[str] = None,
    # Hypothesis-specific configuration
    hypothesis_deadline: Optional[Union[int, NotSet]] = None,
    hypothesis_derandomize: Optional[bool] = None,
    hypothesis_max_examples: Optional[int] = None,
    hypothesis_phases: Optional[List[hypothesis.Phase]] = None,
    hypothesis_report_multiple_bugs: Optional[bool] = None,
    hypothesis_suppress_health_check: Optional[List[hypothesis.HealthCheck]] = None,
    hypothesis_verbosity: Optional[hypothesis.Verbosity] = None,
) -> Generator[events.ExecutionEvent, None, None]:
    """Prepare a generator that will run test cases against the given API definition."""
    # pylint: disable=too-many-locals

    validate_loader(loader, schema_uri)

    if auth is None:
        # Auth type doesn't matter if auth is not passed
        auth_type = None  # type: ignore
    hypothesis_options = prepare_hypothesis_options(
        deadline=hypothesis_deadline,
        derandomize=hypothesis_derandomize,
        max_examples=hypothesis_max_examples,
        phases=hypothesis_phases,
        report_multiple_bugs=hypothesis_report_multiple_bugs,
        suppress_health_check=hypothesis_suppress_health_check,
        verbosity=hypothesis_verbosity,
    )
    return execute_from_schema(
        schema_uri=schema_uri,
        loader=loader,
        base_url=base_url,
        endpoint=endpoint,
        method=method,
        tag=tag,
        operation_id=operation_id,
        app=app,
        validate_schema=validate_schema,
        skip_deprecated_operations=skip_deprecated_operations,
        force_schema_version=force_schema_version,
        checks=checks,
        data_generation_methods=data_generation_methods,
        max_response_time=max_response_time,
        targets=targets,
        hypothesis_options=hypothesis_options,
        seed=seed,
        workers_num=workers_num,
        exit_first=exit_first,
        dry_run=dry_run,
        auth=auth,
        auth_type=auth_type,
        headers=headers,
        request_timeout=request_timeout,
        request_tls_verify=request_tls_verify,
        store_interactions=store_interactions,
        fixups=fixups,
        stateful=stateful,
        stateful_recursion_limit=stateful_recursion_limit,
    )


def validate_loader(loader: Callable, schema_uri: Union[str, Dict[str, Any]]) -> None:
    """Sanity checking for input schema & loader."""
    if loader not in (
        loaders.from_uri,
        loaders.from_aiohttp,
        loaders.from_dict,
        loaders.from_file,
        loaders.from_path,
        loaders.from_wsgi,
    ):
        # Custom loaders are not checked
        return
    if isinstance(schema_uri, dict):
        if loader is not loaders.from_dict:
            raise ValueError("Dictionary as a schema is allowed only with `from_dict` loader")
    elif loader is loaders.from_dict:
        raise ValueError("Schema should be a dictionary for `from_dict` loader")


def execute_from_schema(
    *,
    schema_uri: Union[str, Dict[str, Any]],
    loader: Callable = loaders.from_uri,
    base_url: Optional[str] = None,
    endpoint: Optional[Filter] = None,
    method: Optional[Filter] = None,
    tag: Optional[Filter] = None,
    operation_id: Optional[Filter] = None,
    app: Optional[str] = None,
    validate_schema: bool = True,
    skip_deprecated_operations: bool = False,
    force_schema_version: Optional[str] = None,
    checks: Iterable[CheckFunction],
    data_generation_methods: Tuple[DataGenerationMethod, ...] = DEFAULT_DATA_GENERATION_METHODS,
    max_response_time: Optional[int] = None,
    targets: Iterable[Target],
    workers_num: int = 1,
    hypothesis_options: Dict[str, Any],
    auth: Optional[RawAuth] = None,
    auth_type: Optional[str] = None,
    headers: Optional[Dict[str, Any]] = None,
    request_timeout: Optional[int] = None,
    request_tls_verify: Union[bool, str] = True,
    seed: Optional[int] = None,
    exit_first: bool = False,
    dry_run: bool = False,
    store_interactions: bool = False,
    fixups: Iterable[str] = (),
    stateful: Optional[Stateful] = None,
    stateful_recursion_limit: int = DEFAULT_STATEFUL_RECURSION_LIMIT,
) -> Generator[events.ExecutionEvent, None, None]:
    """Execute tests for the given schema.

    Provides the main testing loop and preparation step.
    """
    # pylint: disable=too-many-locals
    try:
        if fixups:
            if "all" in fixups:
                _fixups.install()
            else:
                _fixups.install(fixups)

        if app is not None:
            app = import_app(app)
        schema = load_schema(
            schema_uri,
            base_url=base_url,
            loader=loader,
            app=app,
            validate_schema=validate_schema,
            skip_deprecated_operations=skip_deprecated_operations,
            auth=auth,
            auth_type=auth_type,
            headers=headers,
            endpoint=endpoint,
            method=method,
            tag=tag,
            operation_id=operation_id,
            data_generation_methods=data_generation_methods,
            force_schema_version=force_schema_version,
            request_tls_verify=request_tls_verify,
        )

        runner: BaseRunner
        if workers_num > 1:
            if not schema.app:
                runner = ThreadPoolRunner(
                    schema=schema,
                    checks=checks,
                    max_response_time=max_response_time,
                    targets=targets,
                    hypothesis_settings=hypothesis_options,
                    auth=auth,
                    auth_type=auth_type,
                    headers=headers,
                    seed=seed,
                    workers_num=workers_num,
                    request_timeout=request_timeout,
                    request_tls_verify=request_tls_verify,
                    exit_first=exit_first,
                    dry_run=dry_run,
                    store_interactions=store_interactions,
                    stateful=stateful,
                    stateful_recursion_limit=stateful_recursion_limit,
                )
            elif isinstance(schema.app, Starlette):
                runner = ThreadPoolASGIRunner(
                    schema=schema,
                    checks=checks,
                    max_response_time=max_response_time,
                    targets=targets,
                    hypothesis_settings=hypothesis_options,
                    auth=auth,
                    auth_type=auth_type,
                    headers=headers,
                    seed=seed,
                    exit_first=exit_first,
                    dry_run=dry_run,
                    store_interactions=store_interactions,
                    stateful=stateful,
                    stateful_recursion_limit=stateful_recursion_limit,
                )

            else:
                runner = ThreadPoolWSGIRunner(
                    schema=schema,
                    checks=checks,
                    max_response_time=max_response_time,
                    targets=targets,
                    hypothesis_settings=hypothesis_options,
                    auth=auth,
                    auth_type=auth_type,
                    headers=headers,
                    seed=seed,
                    workers_num=workers_num,
                    exit_first=exit_first,
                    dry_run=dry_run,
                    store_interactions=store_interactions,
                    stateful=stateful,
                    stateful_recursion_limit=stateful_recursion_limit,
                )

        else:
            if not schema.app:
                runner = SingleThreadRunner(
                    schema=schema,
                    checks=checks,
                    max_response_time=max_response_time,
                    targets=targets,
                    hypothesis_settings=hypothesis_options,
                    auth=auth,
                    auth_type=auth_type,
                    headers=headers,
                    seed=seed,
                    request_timeout=request_timeout,
                    request_tls_verify=request_tls_verify,
                    exit_first=exit_first,
                    dry_run=dry_run,
                    store_interactions=store_interactions,
                    stateful=stateful,
                    stateful_recursion_limit=stateful_recursion_limit,
                )
            elif isinstance(schema.app, Starlette):
                runner = SingleThreadASGIRunner(
                    schema=schema,
                    checks=checks,
                    max_response_time=max_response_time,
                    targets=targets,
                    hypothesis_settings=hypothesis_options,
                    auth=auth,
                    auth_type=auth_type,
                    headers=headers,
                    seed=seed,
                    exit_first=exit_first,
                    dry_run=dry_run,
                    store_interactions=store_interactions,
                    stateful=stateful,
                    stateful_recursion_limit=stateful_recursion_limit,
                )
            else:
                runner = SingleThreadWSGIRunner(
                    schema=schema,
                    checks=checks,
                    max_response_time=max_response_time,
                    targets=targets,
                    hypothesis_settings=hypothesis_options,
                    auth=auth,
                    auth_type=auth_type,
                    headers=headers,
                    seed=seed,
                    exit_first=exit_first,
                    dry_run=dry_run,
                    store_interactions=store_interactions,
                    stateful=stateful,
                    stateful_recursion_limit=stateful_recursion_limit,
                )

        yield from runner.execute()
    except Exception as exc:
        yield events.InternalError.from_exc(exc)


def load_schema(
    schema_uri: Union[str, Dict[str, Any]],
    *,
    base_url: Optional[str] = None,
    loader: Callable = loaders.from_uri,
    app: Any = None,
    validate_schema: bool = True,
    skip_deprecated_operations: bool = False,
    data_generation_methods: Tuple[DataGenerationMethod, ...] = DEFAULT_DATA_GENERATION_METHODS,
    force_schema_version: Optional[str] = None,
    request_tls_verify: Union[bool, str] = True,
    # Network request parameters
    auth: Optional[Tuple[str, str]] = None,
    auth_type: Optional[str] = None,
    headers: Optional[Dict[str, str]] = None,
    # Schema filters
    endpoint: Optional[Filter] = None,
    method: Optional[Filter] = None,
    tag: Optional[Filter] = None,
    operation_id: Optional[Filter] = None,
) -> BaseSchema:
    """Load schema via specified loader and parameters."""
    loader_options = dict_true_values(
        base_url=base_url,
        endpoint=endpoint,
        method=method,
        tag=tag,
        operation_id=operation_id,
        app=app,
        data_generation_methods=data_generation_methods,
    )

    if not isinstance(schema_uri, dict):
        if file_exists(schema_uri):
            loader = loaders.from_path
        elif app is not None and not urlparse(schema_uri).netloc:
            # If `schema` is not an existing filesystem path, or a URL then it is considered as a path within
            # the given app
            loader = loaders.get_loader_for_app(app)
            loader_options.update(dict_true_values(headers=headers))
        else:
            loader_options.update(dict_true_values(headers=headers, auth=auth, auth_type=auth_type))

    if loader is loaders.from_uri and loader_options.get("auth"):
        loader_options["auth"] = get_requests_auth(loader_options["auth"], loader_options.pop("auth_type", None))
    if loader in (loaders.from_uri, loaders.from_aiohttp):
        loader_options["verify"] = request_tls_verify

    return loader(
        schema_uri,
        validate_schema=validate_schema,
        skip_deprecated_operations=skip_deprecated_operations,
        force_schema_version=force_schema_version,
        **loader_options,
    )


def prepare_hypothesis_options(
    deadline: Optional[Union[int, NotSet]] = None,
    derandomize: Optional[bool] = None,
    max_examples: Optional[int] = None,
    phases: Optional[List[hypothesis.Phase]] = None,
    report_multiple_bugs: Optional[bool] = None,
    suppress_health_check: Optional[List[hypothesis.HealthCheck]] = None,
    verbosity: Optional[hypothesis.Verbosity] = None,
) -> Dict[str, Any]:
    options = dict_not_none_values(
        derandomize=derandomize,
        max_examples=max_examples,
        phases=phases,
        report_multiple_bugs=report_multiple_bugs,
        suppress_health_check=suppress_health_check,
        verbosity=verbosity,
    )
    # `deadline` is special, since Hypothesis allows passing `None`
    if deadline is not None:
        if isinstance(deadline, NotSet):
            options["deadline"] = None
        else:
            options["deadline"] = deadline
    return options



import asyncio
from typing import Optional

from aiohttp import web

from . import _server


def _run_server(app: web.Application, port: int) -> None:
    """Run the given app on the given port.

    Intended to be called as a target for a separate thread.
    NOTE. `aiohttp.web.run_app` works only in the main thread and can't be used here (or maybe can we some tuning)
    """
    # Set a loop for a new thread (there is no by default for non-main threads)
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    runner = web.AppRunner(app)
    loop.run_until_complete(runner.setup())
    site = web.TCPSite(runner, "127.0.0.1", port)
    loop.run_until_complete(site.start())
    loop.run_forever()


def run_server(app: web.Application, port: Optional[int] = None, timeout: float = 0.05) -> int:
    """Start a thread with the given aiohttp application."""
    return _server.run(_run_server, app=app, port=port, timeout=timeout)



class RuntimeExpressionError(ValueError):
    """Generic error that happened during evaluation of a runtime expression."""


class UnknownToken(RuntimeExpressionError):
    """Don't know how to handle a token value."""



import os
import shutil
from typing import List, Optional

import attr

from ..runner.serialization import SerializedTestResult


@attr.s(slots=True)  # pragma: no mutate
class ExecutionContext:
    """Storage for the current context of the execution."""

    hypothesis_output: List[str] = attr.ib(factory=list)  # pragma: no mutate
    workers_num: int = attr.ib(default=1)  # pragma: no mutate
    show_errors_tracebacks: bool = attr.ib(default=False)  # pragma: no mutate
    validate_schema: bool = attr.ib(default=True)  # pragma: no mutate
    operations_processed: int = attr.ib(default=0)  # pragma: no mutate
    # It is set in runtime, from a `Initialized` event
    operations_count: Optional[int] = attr.ib(default=None)  # pragma: no mutate
    current_line_length: int = attr.ib(default=0)  # pragma: no mutate
    terminal_size: os.terminal_size = attr.ib(factory=shutil.get_terminal_size)  # pragma: no mutate
    results: List[SerializedTestResult] = attr.ib(factory=list)  # pragma: no mutate
    cassette_file_name: Optional[str] = attr.ib(default=None)  # pragma: no mutate
    junit_xml_file: Optional[str] = attr.ib(default=None)  # pragma: no mutate
    verbosity: int = attr.ib(default=0)  # pragma: no mutate



from typing import Optional

from flask import Flask

from . import _server


def run_server(app: Flask, port: Optional[int] = None, timeout: float = 0.05) -> int:
    """Start a thread with the given aiohttp application."""
    return _server.run(app.run, port=port, timeout=timeout)



from copy import deepcopy
from typing import Any, Dict

from ...utils import traverse_schema


def to_json_schema(schema: Dict[str, Any], nullable_name: str) -> Dict[str, Any]:
    """Convert Open API parameters to JSON Schema.

    NOTE. This function is applied to all keywords (including nested) during a schema resolving, thus it is not recursive.
    See a recursive version below.
    """
    schema = deepcopy(schema)
    if schema.get(nullable_name) is True:
        del schema[nullable_name]
        schema = {"anyOf": [schema, {"type": "null"}]}
    if schema.get("type") == "file":
        schema["type"] = "string"
        schema["format"] = "binary"
    return schema


def to_json_schema_recursive(schema: Dict[str, Any], nullable_name: str) -> Dict[str, Any]:
    return traverse_schema(schema, to_json_schema, nullable_name)



"""Processing of ``securityDefinitions`` or ``securitySchemes`` keywords."""
from typing import Any, ClassVar, Dict, Generator, List, Tuple, Type

import attr
from jsonschema import RefResolver

from ...models import APIOperation
from .parameters import OpenAPI20Parameter, OpenAPI30Parameter, OpenAPIParameter


@attr.s(slots=True)  # pragma: no mutate
class BaseSecurityProcessor:
    api_key_locations: Tuple[str, ...] = ("header", "query")
    http_security_name = "basic"
    parameter_cls: ClassVar[Type[OpenAPIParameter]] = OpenAPI20Parameter

    def process_definitions(self, schema: Dict[str, Any], operation: APIOperation, resolver: RefResolver) -> None:
        """Add relevant security parameters to data generation."""
        for definition in self._get_active_definitions(schema, operation, resolver):
            if definition["type"] == "apiKey":
                self.process_api_key_security_definition(definition, operation)
            self.process_http_security_definition(definition, operation)

    def _get_active_definitions(
        self, schema: Dict[str, Any], operation: APIOperation, resolver: RefResolver
    ) -> Generator[Dict[str, Any], None, None]:
        """Get only security definitions active for the given API operation."""
        definitions = self.get_security_definitions(schema, resolver)
        requirements = get_security_requirements(schema, operation)
        for name, definition in definitions.items():
            if name in requirements:
                yield definition

    def get_security_definitions(self, schema: Dict[str, Any], resolver: RefResolver) -> Dict[str, Any]:
        return schema.get("securityDefinitions", {})

    def get_security_definitions_as_parameters(
        self, schema: Dict[str, Any], operation: APIOperation, resolver: RefResolver, location: str
    ) -> List[Dict[str, Any]]:
        """Security definitions converted to OAS parameters.

        We need it to get proper serialization that will be applied on generated values. For this case it is only
        coercing to a string.
        """
        return [
            self._to_parameter(definition)
            for definition in self._get_active_definitions(schema, operation, resolver)
            if self._is_match(definition, location)
        ]

    def process_api_key_security_definition(self, definition: Dict[str, Any], operation: APIOperation) -> None:
        parameter = self.parameter_cls(self._make_api_key_parameter(definition))
        operation.add_parameter(parameter)

    def process_http_security_definition(self, definition: Dict[str, Any], operation: APIOperation) -> None:
        if definition["type"] == self.http_security_name:
            parameter = self.parameter_cls(self._make_http_auth_parameter(definition))
            operation.add_parameter(parameter)

    def _is_match(self, definition: Dict[str, Any], location: str) -> bool:
        return (definition["type"] == "apiKey" and location in self.api_key_locations) or (
            definition["type"] == self.http_security_name and location == "header"
        )

    def _to_parameter(self, definition: Dict[str, Any]) -> Dict[str, Any]:
        func = {
            "apiKey": self._make_api_key_parameter,
            self.http_security_name: self._make_http_auth_parameter,
        }[definition["type"]]
        return func(definition)

    def _make_http_auth_parameter(self, definition: Dict[str, Any]) -> Dict[str, Any]:
        schema = make_auth_header_schema(definition)
        return make_auth_header(**schema)

    def _make_api_key_parameter(self, definition: Dict[str, Any]) -> Dict[str, Any]:
        return make_api_key_schema(definition, type="string")


def make_auth_header_schema(definition: Dict[str, Any]) -> Dict[str, str]:
    schema = definition.get("scheme", "basic").lower()
    return {"type": "string", "format": f"_{schema}_auth"}


def make_auth_header(**kwargs: Any) -> Dict[str, Any]:
    return {"name": "Authorization", "in": "header", "required": True, **kwargs}


def make_api_key_schema(definition: Dict[str, Any], **kwargs: Any) -> Dict[str, Any]:
    return {"name": definition["name"], "required": True, "in": definition["in"], **kwargs}


SwaggerSecurityProcessor = BaseSecurityProcessor


@attr.s(slots=True)  # pragma: no mutate
class OpenAPISecurityProcessor(BaseSecurityProcessor):
    api_key_locations = ("header", "cookie", "query")
    http_security_name = "http"
    parameter_cls: ClassVar[Type[OpenAPIParameter]] = OpenAPI30Parameter

    def get_security_definitions(self, schema: Dict[str, Any], resolver: RefResolver) -> Dict[str, Any]:
        """In Open API 3 security definitions are located in ``components`` and may have references inside."""
        components = schema.get("components", {})
        security_schemes = components.get("securitySchemes", {})
        if "$ref" in security_schemes:
            return resolver.resolve(security_schemes["$ref"])[1]
        return security_schemes

    def _make_http_auth_parameter(self, definition: Dict[str, Any]) -> Dict[str, Any]:
        schema = make_auth_header_schema(definition)
        return make_auth_header(schema=schema)

    def _make_api_key_parameter(self, definition: Dict[str, Any]) -> Dict[str, Any]:
        return make_api_key_schema(definition, schema={"type": "string"})


def get_security_requirements(schema: Dict[str, Any], operation: APIOperation) -> List[str]:
    """Get applied security requirements for the given API operation."""
    # https://github.com/OAI/OpenAPI-Specification/blob/master/versions/2.0.md#operation-object
    # > This definition overrides any declared top-level security.
    # > To remove a top-level security declaration, an empty array can be used.
    global_requirements = schema.get("security", [])
    local_requirements = operation.definition.raw.get("security", None)
    if local_requirements is not None:
        requirements = local_requirements
    else:
        requirements = global_requirements
    return [key for requirement in requirements for key in requirement]



import platform
from typing import List, Optional

import attr
from click.utils import LazyFile
from junit_xml import TestCase, TestSuite, to_xml_report_file

from ..models import Status
from ..runner import events
from .handlers import EventHandler, ExecutionContext, get_unique_failures


@attr.s(slots=True)  # pragma: no mutate
class JunitXMLHandler(EventHandler):
    file_handle: LazyFile = attr.ib()  # pragma: no mutate
    test_cases: List = attr.ib(factory=list)  # pragma: no mutate
    start_time: Optional[float] = attr.ib(default=None)  # pragma: no mutate

    def handle_event(self, context: ExecutionContext, event: events.ExecutionEvent) -> None:
        if isinstance(event, events.Initialized):
            self.start_time = event.start_time
        if isinstance(event, events.AfterExecution):
            test_case = TestCase(
                f"{event.result.method} {event.result.path}",
                elapsed_sec=event.elapsed_time,
                allow_multiple_subelements=True,
            )
            if event.status == Status.failure:
                checks = get_unique_failures(event.result.checks)
                for idx, check in enumerate(checks, 1):
                    # `check.message` is always not empty for events with `failure` status
                    test_case.add_failure_info(message=f"{idx}. {check.message}")
            if event.status == Status.error:
                test_case.add_error_info(
                    message=event.result.errors[-1].exception, output=event.result.errors[-1].exception_with_traceback
                )
            self.test_cases.append(test_case)
        if isinstance(event, events.Finished):
            test_suites = [TestSuite("schemathesis", test_cases=self.test_cases, hostname=platform.node())]
            to_xml_report_file(file_descriptor=self.file_handle, test_suites=test_suites, prettyprint=True)



"""Expression nodes description and evaluation logic."""
import json
from enum import Enum, unique
from typing import Any, Dict, Optional, Union

import attr
from requests.structures import CaseInsensitiveDict

from ....utils import WSGIResponse
from . import pointers
from .context import ExpressionContext
from .errors import RuntimeExpressionError


@attr.s(slots=True)  # pragma: no mutate
class Node:
    """Generic expression node."""

    def evaluate(self, context: ExpressionContext) -> str:
        raise NotImplementedError


@unique
class NodeType(Enum):
    URL = "$url"
    METHOD = "$method"
    STATUS_CODE = "$statusCode"
    REQUEST = "$request"
    RESPONSE = "$response"


@attr.s(slots=True)  # pragma: no mutate
class String(Node):
    """A simple string that is not evaluated somehow specifically."""

    value: str = attr.ib()  # pragma: no mutate

    def evaluate(self, context: ExpressionContext) -> str:
        """String tokens are passed as they are.

        ``foo{$request.path.id}``

        "foo" is String token there.
        """
        return self.value


@attr.s(slots=True)  # pragma: no mutate
class URL(Node):
    """A node for `$url` expression."""

    def evaluate(self, context: ExpressionContext) -> str:
        return context.case.get_full_url()


@attr.s(slots=True)  # pragma: no mutate
class Method(Node):
    """A node for `$method` expression."""

    def evaluate(self, context: ExpressionContext) -> str:
        return context.case.operation.method.upper()


@attr.s(slots=True)  # pragma: no mutate
class StatusCode(Node):
    """A node for `$statusCode` expression."""

    def evaluate(self, context: ExpressionContext) -> str:
        return str(context.response.status_code)


@attr.s(slots=True)  # pragma: no mutate
class NonBodyRequest(Node):
    """A node for `$request` expressions where location is not `body`."""

    location: str = attr.ib()  # pragma: no mutate
    parameter: str = attr.ib()  # pragma: no mutate

    def evaluate(self, context: ExpressionContext) -> str:
        container: Union[Dict, CaseInsensitiveDict] = {
            "query": context.case.query,
            "path": context.case.path_parameters,
            "header": context.case.headers,
        }[self.location] or {}
        if self.location == "header":
            container = CaseInsensitiveDict(container)
        return str(container[self.parameter])


@attr.s(slots=True)  # pragma: no mutate
class BodyRequest(Node):
    """A node for `$request` expressions where location is `body`."""

    pointer: Optional[str] = attr.ib(default=None)  # pragma: no mutate

    def evaluate(self, context: ExpressionContext) -> Any:
        if self.pointer is None:
            try:
                return json.dumps(context.case.body)
            except TypeError as exc:
                raise RuntimeExpressionError("The request body is not JSON-serializable") from exc
        document = context.case.body
        return pointers.resolve(document, self.pointer[1:])


@attr.s(slots=True)  # pragma: no mutate
class HeaderResponse(Node):
    """A node for `$response.header` expressions."""

    parameter: str = attr.ib()  # pragma: no mutate

    def evaluate(self, context: ExpressionContext) -> str:
        return context.response.headers[self.parameter]


@attr.s(slots=True)  # pragma: no mutate
class BodyResponse(Node):
    """A node for `$response.body` expressions."""

    pointer: Optional[str] = attr.ib(default=None)  # pragma: no mutate

    def evaluate(self, context: ExpressionContext) -> Any:
        if self.pointer is None:
            return context.response.text
        if isinstance(context.response, WSGIResponse):
            document = context.response.json
        else:
            document = context.response.json()
        return pointers.resolve(document, self.pointer[1:])
